#include "server.h" 

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

static OggVorbis_File vorbis_file;
static ogg_page page;
static ogg_stream_state out_state;
static int repos_request = 0;

struct vcut_packet{
        int length;
        unsigned char *packet;
};

struct vcut_state {              
        ogg_sync_state *sync_in;       
        ogg_stream_state *stream_in;   
        ogg_sync_state *sync_out;       
        ogg_stream_state *stream_out;   
        vorbis_dsp_state *vd; 
        vorbis_block *vb;     
        vorbis_info *vi;
        int time;             
        unsigned int serial;  
        struct vcut_packet **headers; /* 3 */ 
        struct vcut_packet **packets; /* 2 */ 

        FILE *in; 
};                 

static struct vcut_state *vcutstate;
static int sent_headers, offset_granulepos;
static ogg_int64_t granpos;
static int have_page, first_page, have_packet, packetno, short_write; 

static void clear_vcut(void)
{
	memset(vcutstate->headers, 0, sizeof(struct vcut_packet)*3);
	memset(vcutstate->packets, 0, sizeof(struct vcut_packet)*2);
}

int ogg_get_file_info(FILE *file, char *info_str, long unsigned *frames, 
	int *seconds)
{
	int ret; 
	vorbis_info *vi;
	
	clear_vcut();
	if (!file) {
		clog(CRIT, "%s: invalid file pointer\n", __func__);
		return -1;
	}
	clog(DEBUG, "%s: ov_open\n", __func__);
	rewind(file);
	ret = ov_open(file, &vorbis_file, NULL, 0);
	if (ret < 0) {
		clog(INFO, "%s: ov_open failed\n", __func__);
		ov_clear(&vorbis_file);
		return ret;
	}
	vi = ov_info(&vorbis_file, 0);
	
	*frames = ov_pcm_total(&vorbis_file, -1);
	*seconds = ov_time_total(&vorbis_file, -1);
	sprintf(info_str, "audio_file_info1:%ldkHz, %d channels, %ldkbps, "
		"%lu samples\n"
		"audio_file_info2: \n"
		"audio_file_info3: \n",
		vi->rate / 1000 , vi->channels, ov_bitrate(&vorbis_file, 0) / 1000,
		*frames);
	clog(WARNING, "%s: success\n", __func__);
	return 1;
}

static struct vcut_packet *save_packet(ogg_packet *packet)
{
        struct vcut_packet *p = malloc(sizeof(struct vcut_packet));

        p->length = packet->bytes;
        p->packet = malloc(p->length);
        memcpy(p->packet, packet->packet, p->length);

        return p;
}

/* Pull out and save the 3 header packets from the input file.
 */     
static int process_headers(struct vcut_state *s)
{              
        vorbis_comment vc;    
        ogg_page page;
        ogg_packet packet;
        int bytes;
        int i;
        unsigned char *buffer;
        
	clog(DEBUG, "%s: \n", __func__);
        ogg_sync_init(s->sync_in);

        vorbis_info_init(s->vi);
        vorbis_comment_init(&vc);

        buffer = ogg_sync_buffer(s->sync_in, 4096);
        bytes = fread(buffer, 1, 4096, s->in);
        ogg_sync_wrote(s->sync_in, bytes);

	if (ogg_sync_pageout(s->sync_in, &page) != 1){
		clog(ERROR, "%s: input not ogg.\n", __func__);
		return -1;
	}

        s->serial = ogg_page_serialno(&page);
	ogg_stream_init(s->stream_in, s->serial);

        if (ogg_stream_pagein(s->stream_in, &page) < 0) {
		clog(ERROR, "%s: error in first page\n", __func__);
		return -1;
        }

	if (ogg_stream_packetout(s->stream_in, &packet) != 1) {
		clog(ERROR, "%s: error in first packet\n", __func__);
		return -1;
	}

        if (vorbis_synthesis_headerin(s->vi, &vc, &packet) < 0) {       
		clog(ERROR, "%s: error in primary header: not vorbis?\n", 
			__func__);
		return -1;
        }
	s->headers[0] = save_packet(&packet);
        i = 0;
        while (i < 2) {       
		while (i < 2) {
			int res = ogg_sync_pageout(s->sync_in, &page);
			if(res == 0)
				break;
			if (res == 1) {
				ogg_stream_pagein(s->stream_in, &page);
				while (i < 2) {
					res = ogg_stream_packetout(
						s->stream_in, &packet);
					if (res == 0)	
						break;
					if (res < 0) {
						clog(ERROR, "%s: Secondary "
							"header corrupt\n", 
							__func__);
						return -1;
					}
					s->headers[i + 1] = 
						save_packet(&packet);
					vorbis_synthesis_headerin(s->vi, 
						&vc, &packet);
					i++;
				}
			}
		}
		buffer = ogg_sync_buffer(s->sync_in, 4096);
		bytes = fread(buffer, 1, 4096, s->in);
		if (bytes == 0 && i < 2) {
			clog(ERROR, "%s: EOF in headers\n", __func__);
			return -1;
		}
		ogg_sync_wrote(s->sync_in, bytes);
	}
        vorbis_comment_clear(&vc);
	clog(DEBUG, "%s: success\n", __func__);
        return 0;
}


static struct vcut_state *vcut_new(void)
{
	struct vcut_state *s = malloc(sizeof(struct vcut_state));
	memset(s,0,sizeof(struct vcut_state));

	s->sync_in = malloc(sizeof(ogg_sync_state));
	s->stream_in = malloc(sizeof(ogg_stream_state));
	s->vd = malloc(sizeof(vorbis_dsp_state));
	s->vi = malloc(sizeof(vorbis_info)); 
	s->vb = malloc(sizeof(vorbis_block));  
	s->headers = malloc(sizeof(struct vcut_packet)*3);
	s->packets = malloc(sizeof(struct vcut_packet)*2);
	vcutstate = s;
	clear_vcut();
	return s;
}

static void submit_headers_to_stream(ogg_stream_state *state)
{
	struct vcut_state *s = vcutstate;
        int i;

	clog(DEBUG, "%s: submitting three headers\n", __func__);
	for (i = 0; i < 3; i++) {
		ogg_packet p;
		p.bytes = s->headers[i]->length;
		p.packet = s->headers[i]->packet;
		p.b_o_s = ((i == 0)? 1 : 0);        
		p.e_o_s = 0;
		p.granulepos = 0;
		ogg_stream_packetin(state, &p);
	}
}

static int update_sync(struct vcut_state *s, FILE *f)
{
        unsigned char *buffer = ogg_sync_buffer(s->sync_in, 4096);
        int bytes = fread(buffer, 1, 4096, f);
        ogg_sync_wrote(s->sync_in, bytes);
        return bytes;
}

static void ogg_close_audio_file(FILE* audio_file)
{
	//close_stream_writer_pipe(stream_writer);
	//kill_stream_writer();
	if (audio_file) {
		clog(DEBUG, "%s: ov_clear\n", __func__);
		ov_clear(&vorbis_file);
	}
	clear_vcut();
	repos_request = 0;
	have_page = 0;
	have_packet = 0;
	sent_headers = 0;
	packetno = 0;
	offset_granulepos = 0;
}
static int ogg_reposition_stream(FILE* file, long unsigned pcm_no, pid_t *pid)
{
	
	clog(DEBUG, "%s: preparing jump to %lu\n", __func__, pcm_no);
	if (!vcutstate)
		return -1;
	//close_stream_writer_pipe();
	//kill_stream_writer();
	sent_headers = 0;
	have_page = 0;
	have_packet = 0;
	//sleep(1);
	repos_request = pcm_no;
	return 0;
}


static int page_out(int flush)
{
	int (*fn)(ogg_stream_state *, ogg_page *) = flush? 
		ogg_stream_flush : ogg_stream_pageout; 
	return fn(&out_state, &page);
}


static int ogg_write_chunk(FILE *file, char *info_str, long unsigned *num_pcm)
{
	ogg_page *p = &page;
	static ssize_t len;
	static char *buf;
	int ret = -1;

	if (!short_write) {
		len = p->body_len + p->header_len;
		buf = malloc(len + 1);
		memcpy(buf, p->header, p->header_len);
		memcpy(buf + p->header_len, p->body, p->body_len);
	}
	if (!buf)
		goto out;
	ret = write(fileno(file), buf, len);
	if (ret < 0)
		clog(ERROR, "%s: can not write to fd %d (%s)\n", __func__, fileno(file),
			strerror(errno));
	if (ret <= 0)
		goto out;
	//clog(DEBUG, "%s: updating chunks_sent: %d\n", __func__, *num_pcm);
	if (!repos_request && !first_page)
		*num_pcm = granpos + offset_granulepos;
	if (ret < len) {
		//clog(DEBUG, "%s: short write (%i/%i)\n", __func__, ret, p->body_len);
		memcpy(buf, buf + ret, p->body_len + p->header_len - ret);
		len -= ret;
		short_write = 1;
		return 1;
	}
	short_write = 0;
	//clog(DEBUG, "%s: %llu\n", "__func__", ogg_pages_sent);
out:
	clear_vcut();	
	if (buf)
		free(buf);
	return ret;
}


static int get_page(FILE *audio_file)
{
	int ret;

	for (;;) {
	//	clog(DEBUG, "%s: sync_pageout\n", __func__);
		ret = ogg_sync_pageout(vcutstate->sync_in, &page);
		if (ret < 0)
			continue;
		if (!ret) {
			update_sync(vcutstate, audio_file);
			continue;
		}
//		clog(DEBUG, "%s: page in\n", __func__);
		ret = ogg_stream_pagein(vcutstate->stream_in, &page);
		if (ret < 0) {
			clog(DEBUG, "%s: stream_pagein failed\n", __func__);
			return -1;
		}
		if (ogg_page_eos(&page))
			return 0;
		return 1;
	}
}

static int ogg_read_chunk(FILE *audio_file)
{
	struct vcut_state *s = vcutstate;
	int ret;
	ogg_packet packet;
	ogg_packet dummy;
	static int packet_out_failed;
	if (short_write)
		return 1;	
	//clog(DEBUG, "%s: \n", __func__);
	if (!sent_headers) {
		clear_vcut();
		rewind(audio_file);
		s->in = audio_file;
		if (process_headers(vcutstate) < 0)
			return -1;
		clog(DEBUG, "%s: synthesis_init\n", __func__);
	        vorbis_synthesis_init(s->vd, s->vi);
		clog(DEBUG, "%s: block_init\n", __func__);
		vorbis_block_init(s->vd, s->vb);
		clog(DEBUG, "%s: stream_init\n", __func__);
		ogg_stream_init(&out_state, s->serial);
		clog(DEBUG, "%s: submit headers\n");
		submit_headers_to_stream(&out_state);
		first_page = 1;
		have_page = 1;
		have_packet = 1;
		sent_headers = 1;
		return page_out(1)? 1 : -1;
	}
get_page:
	if (repos_request && !first_page && !have_page) {
		clear_vcut();
		rewind(audio_file);
		ret = ov_pcm_seek_page(&vorbis_file, repos_request);
		clog(DEBUG, "%s: ov_seek_page returned %d\n", 
			__func__, ret);
again:
		get_page(audio_file);
		ret = get_page(audio_file);
		if (ret < 0) {
			clog(ERROR, "%s: pageout failed\n", __func__);
			goto err_out;
		}
		offset_granulepos = ogg_page_granulepos(&page);
		clog(DEBUG, "%s: current granpos: %d\n", 
			__func__, offset_granulepos);
		if (offset_granulepos < repos_request)
			goto again;
		clog(DEBUG, "%s: using offset %d\n", __func__, 
			offset_granulepos);
		packet_out_failed = 2;
		//granpos = offset_granulepos;
		dummy = packet;
		have_page = 1;
		have_packet = 0;
		packetno = 0;
		repos_request = 0;
	}
	if (!have_page) {
		ogg_int64_t tmp;

		ret = get_page(audio_file);
		if (ret <= 0)
			goto err_out;
		have_packet = 0;
		tmp = ogg_page_granulepos(&page);
		if (tmp == -1)
			goto get_page;
		have_page = 1;
		granpos = tmp - offset_granulepos;
	}
	if (!have_packet) {
get_packet:
		ret = ogg_stream_packetout(vcutstate->stream_in, &packet);
		if (ret < 0) {
			clog(ERROR, "%s: packetout failed\n", __func__);
			//packet_out_failed++;
			goto get_packet;
		}
		if (!ret) {
			have_page = 0;
			goto get_page;
		}
		if (vcutstate->vi) {
			dummy = packet;
			vorbis_packet_blocksize(vcutstate->vi, 
				&dummy);
		}
		have_packet = 1;
		packet.packetno = packetno++;
		packet.granulepos = granpos;
		ogg_stream_packetin(&out_state, &packet);
		//clog(DEBUG, "%s: gp: %li\n", __func__, granpos);
		//clog(DEBUG, "%s: no: %d\n", __func__, packetno);
	}
	if (!page_out(first_page)) {
		have_packet = 0;
		if (first_page) {
			first_page = 0;
			have_page = 0;
		}
		goto get_page;
	}
//success:
#if 1
	if (packet_out_failed) {
		have_page = 0;
		packet_out_failed--;
		goto get_page;
	}
#endif
	return 1;
err_out:
	clear_vcut();
	have_page = 0;
	have_packet = 0;
	sent_headers = 0;
	packetno = 0;
	offset_granulepos = 0;
	return ret;
}

void ogg_end_of_input(FILE *in_file, FILE **sw, pid_t *pid)
{
	int status;
	clog(DEBUG, "%s: ogg_close\n", __func__);
	ogg_close_audio_file(in_file);
	//usleep(5000 * 1000);
	if (*pid) {
		kill(*pid, SIGTERM);
		waitpid(*pid, &status, 0);
		//usleep(5000 * 1000);
	}
	if (*sw)
		pclose(*sw);
	*sw = NULL;
	*pid = 0;
	//usleep(5000 * 1000);
	clear_vcut();
	have_page = 0;
	have_packet = 0;
	sent_headers = 0;
	packetno = 0;
	offset_granulepos = 0;
	short_write = 0;
}

int ogg_init(void *p)
{
	struct audio_format *af = p;

	af->get_file_info = ogg_get_file_info,
	af->reposition_stream = ogg_reposition_stream,
	af->read_chunk = ogg_read_chunk;
	af->write_chunk = ogg_write_chunk;
	af->close_audio_file = ogg_close_audio_file;
	vcutstate = vcut_new();
	af->end_of_input = ogg_end_of_input;
	return 1;
}
